import json
from pathlib import Path
from typing import Type, Union

import panel as pn
import param
from panel.custom import AnyWidgetComponent

from oold.model import LinkedBaseModel
from oold.model.v1 import LinkedBaseModel as LinkedBaseModel_v1

pn.extension()

bundled_assets_dir = Path(__file__).parent.parent.parent / "vue" / "dist" / "default"


class JsonEditor(AnyWidgetComponent):
    """A JSON-SCHEMA based form editor using
    https://github.com/json-editor/json-editor"""

    _esm = (bundled_assets_dir / "jsoneditor_vue.mjs").read_text()

    _stylesheets = [
        # includes bootstrap and spectre
        (bundled_assets_dir / "jsoneditor_vue.css").read_text(),
        # v5 does not work properly:
        # "https://cdn.jsdelivr.net/npm/bootstrap@4/dist/css/bootstrap.min.css",
        # does not work:
        # 'https://use.fontawesome.com/releases/v5.12.1/css/all.css',
        # "https://unpkg.com/spectre.css/dist/spectre-icons.min.css",
    ]
    _importmap = {
        "imports": {
            "vue": "https://esm.sh/vue@3",
            # works with `import {JSONEditor} from "@json-editor/json-editor"`:
            # "@json-editor/json-editor": "https://esm.sh/@json-editor/json-editor@latest",  # noqa
            # works with `import("@json-editor/json-editor")`:
            # "@json-editor/json-editor": (
            #   "https://cdn.jsdelivr.net/npm/@json-editor/json-editor",
            #   "@latest/dist/jsoneditor.min.js"
            # ),
            # works with `import("jsoneditor")`:
            "jsoneditor": "https://cdn.jsdelivr.net/npm/@json-editor/json-editor@latest/dist/jsoneditor.min.js",  # noqa
        }
    }
    # __javascript__= [
    #     "https://cdn.jsdelivr.net/npm/vue@2/dist/vue.js",
    #     "https://unpkg.com/bootstrap-vue@latest/dist/bootstrap-vue.min.js",
    #     "https://cdn.jsdelivr.net/npm/@json-editor/json-editor@latest/dist/jsoneditor.min.js"
    # ]
    value = param.Dict()
    options = param.Dict(
        default={
            # "theme": "bootstrap4",
            # "iconlib": 'fontawesome5',
            # "iconlib": "spectre",
            "schema": {
                "required": ["testxy"],
                "properties": {"testxy": {"type": "string"}},
            },
        }
    )
    ready = param.Boolean(default=False, doc="Indicates if the JSONEditor is ready.")

    encoder = param.ClassSelector(
        class_=json.JSONEncoder,
        is_instance=False,
        doc="""
    Custom JSONEncoder class used to serialize objects to JSON string.""",
    )

    def get_value(self):
        json_str = json.dumps(self.value, cls=self.encoder)
        return json.loads(json_str)

    def set_value(self, value: dict):
        """Set the value of the JSON editor."""
        self.value = value

    def set_schema(self, schema: dict, keep_value: bool = True):
        """Set the schema of the JSON editor."""
        # override options param to trigger change event
        new_options = {**self.options, "schema": schema}
        if keep_value:
            new_options["startval"] = self.get_value()
        else:
            new_options["startval"] = None
        self.options = new_options


class OoldEditor(JsonEditor):
    """JsonEditor generated from an oold model schema."""

    def __init__(
        self,
        oold_model: Union[Type[LinkedBaseModel], Type[LinkedBaseModel_v1]] = None,
        **params
    ):
        options = params.get("options", {})

        if oold_model is not None:
            options["schema"] = oold_model.export_schema()
        else:
            options["schema"] = {}

        params["options"] = options
        super().__init__(**params)



# generated by datamodel-codegen:
#   filename:  Foo.json
#   timestamp: 2025-06-16T05:00:41+00:00

from __future__ import annotations

from typing import List, Optional

from pydantic import ConfigDict, Field

from oold.model import LinkedBaseModel


class Bar2(LinkedBaseModel):
    model_config = ConfigDict(
        json_schema_extra={"title": "Bar2"},
    )
    id: Optional[str] = None
    type: Optional[List[str]] = ["Bar2"]
    prop1: Optional[str] = None


class Bar(Bar2):
    model_config = ConfigDict(
        json_schema_extra={"title": "Bar"},
    )
    type: Optional[List[str]] = ["Bar"]
    prop2: Optional[str] = None


class Foo(LinkedBaseModel):
    model_config = ConfigDict(
        json_schema_extra={"title": "Foo"},
    )
    id: str
    type: Optional[List[str]] = ["Foo"]
    literal: Optional[str] = None
    b: Optional[Bar] = Field(
        None, json_schema_extra={"range": "Bar.json", "x-oold-required-iri": True}
    )
    b_default: Optional[Bar] = Field(
        default_factory=lambda: Bar.model_validate("ex:b"),
        json_schema_extra={"range": "Bar.json"},
    )
    b_set_later: Optional[Bar] = Field(None, json_schema_extra={"range": "Bar.json"})
    b2: Optional[List[Bar]] = Field(None, json_schema_extra={"range": "Bar.json"})



# generated by datamodel-codegen:
#   filename:  example.json
#   timestamp: 2025-06-16T04:13:18+00:00

from __future__ import annotations

from typing import Optional, Union

from pydantic import ConfigDict, Field

from oold.model import LinkedBaseModel


class Subschema1(LinkedBaseModel):
    model_config = ConfigDict(
        json_schema_extra={"title": "Subschema1"},
    )
    subprop0: Optional[str] = Field(
        None, json_schema_extra={"custom_key": "custom_value_0"}
    )
    subprop1: Optional[str] = Field(
        None, json_schema_extra={"custom_key": "custom_value_1"}
    )


class Subschema2(LinkedBaseModel):
    model_config = ConfigDict(
        json_schema_extra={"title": "Subschema2"},
    )
    subprop0: Optional[str] = Field(
        None, json_schema_extra={"custom_key": "custom_value_0"}
    )
    subprop2: Optional[str] = Field(
        None, json_schema_extra={"custom_key": "custom_value_2"}
    )


class Example(LinkedBaseModel):
    model_config = ConfigDict(
        json_schema_extra={"title": "Example"},
    )
    type: Optional[str] = ["example"]
    prop1: Optional[str] = Field(None, json_schema_extra={"custom_key": "custom_value"})
    prop2: Optional[Union[Subschema1, Subschema2]] = Field(
        None, json_schema_extra={"custom_key": "custom_value"}
    )



import panel as pn

from oold.ui.panel.anywidget_vite.jsoneditor import JsonEditor


class App(pn.viewable.Viewer):
    def __init__(self, **params):
        super().__init__(**params)

        self.jsoneditor = JsonEditor(max_height=500, max_width=800)

        self.save_btn = pn.widgets.Button(
            css_classes=["save_btn"], name="Save", button_type="primary"
        )
        pn.bind(self.on_save, self.save_btn, watch=True)

        self._view = pn.Column(
            self.jsoneditor,
            pn.pane.JSON(self.jsoneditor.param.value, theme="light"),
            self.save_btn,
        )

    def on_save(self, event):
        # Handle the save event here
        print("Save button clicked")
        print("Current value:", self.jsoneditor.get_value())
        # Update the schema or any other logic as needed
        self.jsoneditor.options = {
            **self.jsoneditor.options,
            "schema": {**self.jsoneditor.options["schema"], "title": "Updated Title"},
        }

    def __panel__(self):
        return self._view


if pn.state.served:
    pn.extension(sizing_mode="stretch_width")

    App().servable()

if __name__ == "__main__":
    pn.serve(App().servable())



from typing import Optional, Type, Union

import panel as pn
from rdflib import Graph

from oold.model import LinkedBaseModel
from oold.model.v1 import LinkedBaseModel as LinkedBaseModel_v1
from oold.ui.panel import JsonEditor

pn.extension("codeeditor")


class OoldDemoEditor(pn.viewable.Viewer):
    def __init__(
        self,
        oold_model: Union[Type[LinkedBaseModel], Type[LinkedBaseModel_v1]] = None,
        options: Optional[dict] = None,
        **params
    ):
        options = params.get("options", {})
        json_editor_options = options.get("json_editor", {})
        if oold_model is not None:
            json_editor_options["schema"] = oold_model.export_schema()

        self._oold_model = oold_model

        super().__init__(**params)

        self.message = pn.pane.Markdown(
            """This is a schema generated editor that
            generates RDF directly from your input"""
        )
        self.jsoneditor = JsonEditor(
            # min_height=500,
            # max_height=500,
            sizing_mode="stretch_width",
            options=json_editor_options,
        )

        self.code_editor = pn.widgets.CodeEditor(
            value="", sizing_mode="stretch_width", language="turtle", height=300
        )

        self.save_btn_clicked = False
        self.save_btn = pn.widgets.Button(
            css_classes=["save_btn"], name="Save", button_type="primary"
        )

        self.jsoneditor.param.watch(self.on_value_change, "value")
        pn.bind(self.on_save, self.save_btn, watch=True)

        self._view = pn.Column(
            self.message,
            pn.Row(self.jsoneditor, scroll=True),
            # display jsoneditor value in a JSON pane for debugging
            # pn.pane.JSON(self.jsoneditor.param.value, theme="light"),
            pn.pane.Markdown(
                "### RDF Representation (Turtle format)", sizing_mode="stretch_width"
            ),
            self.code_editor,
            self.save_btn,
            scroll=True,
            width=800,
        )

    def on_value_change(self, event):
        """Handle changes in the value."""
        # Here you can handle the change event, e.g., save to a database or file
        print("Value changed:", self.jsoneditor.get_value())
        # You can also update the schema or perform other actions based on the new value
        # self.set_schema(self.options["schema"], keep_value=True)

        # construct oold model instance
        instance = None
        try:
            instance = self._oold_model(**self.jsoneditor.get_value())
            print("Valid instance created:", instance)
        except Exception as ex:
            print("Error creating instance:", ex)

        # Update the code editor with the RDF representation
        if instance is not None:
            g = Graph()
            g.parse(data=instance.to_jsonld(), format="json-ld")
            self.code_editor.value = g.serialize(format="turtle")

    def on_save(self, event):
        # Handle the save event here
        print("on_save")
        self.save_btn_clicked = True

    def __panel__(self):
        return self._view



#! pip install nicegui pyscss
from pathlib import Path

from nicegui import app, ui

from oold.model import LinkedBaseModel
from oold.ui.nicegui.jsoneditor import OswEditor


class Item(LinkedBaseModel):
    """A sample item model."""

    name: str = "Sample Item"
    description: str = "This is a sample item description."


def change(data):
    print(data)


# jsoneditor = JsonEditor()
jsoneditor = OswEditor(entity=Item)
ui.button("Clear")  # , on_click=jsoneditor.clear)

# see https://github.com/zauberzeug/nicegui/issues/4177
bundled_assets_dir = Path(__file__).parent.parent / "vue" / "dist" / "default"
print(bundled_assets_dir)
css_name = "bootstrap_scoped.css"
app.add_static_file(
    local_file=bundled_assets_dir / "jsoneditor_vue.css",
    url_path="/" + css_name,
)
ui.add_head_html(
    rf'<link rel="stylesheet" href="/{css_name}">',
    shared=True,
)

ui.run(port=8333)
# ui.run(reload=False, native=True)



from static import _run

from oold.model.static import enum_docstrings as parse_enum_docstrings


def test_oneof_subschema():
    # json schema with property that contains a oneOf with two subschemas

    schemas = [
        {
            "id": "example",
            "title": "Example",
            "type": "object",
            "properties": {
                "type": {"type": "string", "default": ["example"]},
                "prop1": {"type": "string", "custom_key": "custom_value"},
                "prop2": {
                    "custom_key": "custom_value",
                    "properties": {
                        "subprop0": {"type": "string", "custom_key": "custom_value_0"},
                    },
                    "oneOf": [
                        {
                            "title": "Subschema1",
                            "type": "object",
                            "properties": {
                                "subprop1": {
                                    "type": "string",
                                    "custom_key": "custom_value_1",
                                },
                            },
                        },
                        {
                            "title": "Subschema2",
                            "type": "object",
                            "properties": {
                                "subprop2": {
                                    "type": "string",
                                    "custom_key": "custom_value_2",
                                },
                            },
                        },
                    ],
                },
            },
        },
    ]

    def oneof_subschema(pydantic_version):
        # Test the generated model, see
        # https://github.com/koxudaxi/datamodel-code-generator/issues/2403

        if pydantic_version == "v1":
            import data.oneof_subschema.model_v1 as model

            assert (
                model.Subschema1.__fields__["subprop1"].field_info.extra["custom_key"]
                == "custom_value_1"
            )
        else:
            import data.oneof_subschema.model_v2 as model

            model.Subschema1.model_fields["subprop1"].json_schema_extra[
                "custom_key"
            ] == "custom_value_1"

    _run(
        schemas,
        main_schema="example.json",
        test=oneof_subschema,
        # pydantic_versions=["v1"],
    )


def test_enum_docstrings():
    schemas = [
        {
            "id": "example",
            "title": "Example",
            "type": "object",
            "properties": {
                "type": {"type": "string", "default": ["example"]},
                "hobby": {
                    "type": "string",
                    "enum": ["ex:sports", "ex:music", "ex:art"],
                    "description": "Defines various hobbies as an enum.",
                    "x-enum-varnames": ["SPORTS", "MUSIC", "ART"],
                    "options": {
                        "enum_titles": [
                            "Sports hobby, e.g. football, basketball, etc.",
                            "Music hobby, e.g. playing instruments, singing, etc.",
                            "Art hobby, e.g. painting, drawing, etc.",
                        ]
                    },
                },
            },
        },
    ]

    def enum_docstrings(pydantic_version):
        # Test the generated model, see
        # https://github.com/koxudaxi/datamodel-code-generator/issues/2403

        if pydantic_version == "v1":
            import data.enum_docstrings.model_v1 as model

            Hobby = parse_enum_docstrings(model.Hobby)
            assert (
                Hobby.SPORTS.__doc__.strip()
                == "Sports hobby, e.g. football, basketball, etc."
            )
            assert (
                Hobby.MUSIC.__doc__.strip()
                == "Music hobby, e.g. playing instruments, singing, etc."
            )
            assert (
                Hobby.ART.__doc__.strip() == "Art hobby, e.g. painting, drawing, etc."
            )
        else:
            import data.enum_docstrings.model_v2 as model

            Hobby = parse_enum_docstrings(model.Hobby)
            assert (
                Hobby.SPORTS.__doc__.strip()
                == "Sports hobby, e.g. football, basketball, etc."
            )
            assert (
                Hobby.MUSIC.__doc__.strip()
                == "Music hobby, e.g. playing instruments, singing, etc."
            )
            assert (
                Hobby.ART.__doc__.strip() == "Art hobby, e.g. painting, drawing, etc."
            )

    _run(
        schemas,
        main_schema="example.json",
        test=enum_docstrings,
        # pydantic_versions=["v1"],
    )


if __name__ == "__main__":
    test_oneof_subschema()
    test_enum_docstrings()



import json
from typing import Optional

import jsondiff

from oold.model.static import SchemaExportMode


def _run(pydantic_version="v1"):
    if pydantic_version == "v1":
        from pydantic.v1 import BaseModel, Field

        from oold.model.v1 import LinkedBaseModel

        class SubObject(BaseModel):
            prop1: int
            prop2: int

        class MyRootSchema(LinkedBaseModel):
            class Config:
                schema_extra = {
                    "@context": [
                        {"some_property": "https://example.org/some_property"},
                    ]
                }

        class MyCustomSchema(MyRootSchema):
            class Config:
                schema_extra = {
                    "@context": [
                        {"my_property": "https://example.org/my_property"},
                    ],
                }

            a_simple_property: int
            """
            A simple test
            """
            my_property: Optional[str] = Field(
                "default value",
                title="My property",
                title_={"de": "Mein Attribut"},
                description="A test property",
                description_={"de": "Dies ist ein Test-Attribut"},
            )

            a_object_property: SubObject

    else:
        from pydantic import BaseModel, Field

        from oold.model import LinkedBaseModel

        class SubObject(BaseModel):
            prop1: int
            prop2: int

        class MyRootSchema(LinkedBaseModel):
            model_config = {
                "json_schema_extra": {
                    "@context": [
                        {"some_property": "https://example.org/some_property"},
                    ]
                }
            }

        class MyCustomSchema(MyRootSchema):
            model_config = {
                "json_schema_extra": {
                    "@context": [
                        {"my_property": "https://example.org/my_property"},
                    ],
                }
            }

            a_simple_property: int
            """
            A simple test
            """
            my_property: Optional[str] = Field(
                "default value",
                title="My property",
                title_={"de": "Mein Attribut"},
                description="A test property",
                description_={"de": "Dies ist ein Test-Attribut"},
            )

            a_object_property: SubObject

    my_schema = MyCustomSchema.export_schema(mode=SchemaExportMode.PARTIAL)

    print(json.dumps(my_schema, indent=2))
    expected = {
        "@context": [
            "MyRootSchema",
            {"my_property": "https://example.org/my_property"},
        ],
        "$defs": {
            "SubObject": {
                "title": "SubObject",
                "type": "object",
                "properties": {
                    "prop1": {"title": "Prop1", "type": "integer"},
                    "prop2": {"title": "Prop2", "type": "integer"},
                },
                "required": ["prop1", "prop2"],
            }
        },
        "title": "MyCustomSchema",
        "type": "object",
        "properties": {
            "a_simple_property": {"title": "A Simple Property", "type": "integer"},
            "my_property": {
                "title": "My property",
                "description": "A test property",
                "default": "default value",
                "title*": {"de": "Mein Attribut"},
                "description*": {"de": "Dies ist ein Test-Attribut"},
                "type": "string",
            },
            "a_object_property": {"$ref": "#/$defs/SubObject"},
        },
        "required": ["a_simple_property", "a_object_property"],
    }
    diff = jsondiff.diff(my_schema, expected)
    assert diff == {}, f"Schema mismatch for {pydantic_version}:\n{diff}"


def test_schema_generation():
    _run(pydantic_version="v1")
    _run(pydantic_version="v2")


if __name__ == "__main__":
    # Run the test function
    test_schema_generation()



"""Shared configuration and fixtures for testing Panel"""

import panel as pn
import pytest

PORT = [6000]


@pytest.fixture
def port():
    PORT[0] += 1
    return PORT[0]


@pytest.fixture(autouse=True)
def server_cleanup():
    """
    Clean up server state after each test.
    """
    try:
        yield
    finally:
        pn.state.reset()



from pathlib import Path
from typing import Callable, List

import datamodel_code_generator

from oold.generator import Generator


def _run(
    schemas: List, main_schema: str, test: Callable, pydantic_versions=["v1", "v2"]
):
    for pydantic_version in pydantic_versions:
        if pydantic_version == "v1":
            # from oold.model.v1 import (
            #     ResolveParam,
            #     Resolver,
            #     ResolveResult,
            #     SetResolverParam,
            #     set_resolver,
            # )

            output_model_type = datamodel_code_generator.DataModelType.PydanticBaseModel
        else:
            # from oold.model import (
            #     ResolveParam,
            #     Resolver,
            #     ResolveResult,
            #     SetResolverParam,
            #     set_resolver,
            # )

            output_model_type = (
                datamodel_code_generator.DataModelType.PydanticV2BaseModel
            )

        g = Generator()
        g.generate(
            Generator.GenerateParams(
                json_schemas=schemas,
                main_schema=main_schema,
                output_model_type=output_model_type,
                output_model_path=Path(__file__).parent
                / "data"
                / test.__name__
                / ("model_" + pydantic_version + ".py"),
                # working_dir_path=Path(__file__).parent / "data" / "test_core" / "src"
            )
        )

        test(pydantic_version)



import os
from pathlib import Path

import anywidget
import traitlets

bundled_assets_dir = Path(os.path.abspath("")).parent / "vue" / "dist" / "anywidget"


class JsonEditorWidget(anywidget.AnyWidget):
    # _esm = "jsoneditor.mjs" # works
    # _css =  'https://cdn.jsdelivr.net/npm/bootstrap@4.6.1/dist/css/bootstrap.min.css'
    #   works but conflict with vscode
    # _esm = str(bundled_assets_dir / "jsoneditor_vue.mjs")
    # we need to prefix this static import since the dynamic import
    # import "https://esm.sh/@json-editor/json-editor@latest" does not work
    _esm = (
        'import {JSONEditor} from "https://esm.sh/@json-editor/json-editor@latest"\n'
        + (bundled_assets_dir / "jsoneditor_vue.mjs").read_text()
    )
    # scoped css does not conflict with vscode styling
    _css = (bundled_assets_dir / "jsoneditor_vue.css").read_text()

    value = traitlets.Dict({}).tag(sync=True)
    options = traitlets.Dict({}).tag(sync=True)

    @traitlets.observe("value")
    def _observe_value(self, change):
        pass
        # changes can be handled here
        # print(change["old"])
        # print(change["new"])



# pip install panel pytest pytest-playwright
# playwright install
# pytest panel_frontend_test.py --headed --slowmo 1000

import time

import panel as pn
import pytest
from playwright.sync_api import Page

# pytest ./tests/ui/panel/panel_frontend_test.py --headed --slowmo 1000


# required playwright installed, skip for now
@pytest.mark.skip(reason="Playwright tests do not run in CI")
def test_component(page: Page, port):
    from panel_app import App

    app = App()
    url = f"http://localhost:{port}"

    server = pn.serve(app, port=port, threaded=True, show=False)
    time.sleep(0.2)

    page.goto(url)
    time.sleep(3)  # wait for page to load

    # print(json.dumps(app.jsoneditor.get_value()))
    assert app.jsoneditor.get_value() == {"testxy": ""}

    # note: css selector for id=root[testxy] needs to escaped,
    # see https://stackoverflow.com/questions/1466103/escape-square-brackets-when-assigning-a-class-name-to-an-element # noqa
    page.locator("#root\\[testxy\\]").fill("test123")
    page.locator("[for=root\\[testxy\\]]").click()
    assert app.jsoneditor.get_value() == {"testxy": "test123"}

    # click save button (button element with text "Save")
    page.locator("button:has-text('Save')").click()

    time.sleep(0.5)  # wait for save action to complete
    # check if the span with class "je-object__title" contains "Updated Title"
    assert page.locator(".je-object__title").text_content() == "Updated Title"

    assert page.locator("#root\\[testxy\\]").input_value() == "test123"

    app.jsoneditor.set_value({"testxy": "new value"})
    time.sleep(0.5)
    # assert that the input field has the value "new value"
    assert page.locator("#root\\[testxy\\]").input_value() == "new value"

    server.stop()



# generated by datamodel-codegen:
#   filename:  example.json
#   timestamp: 2025-06-16T04:13:17+00:00

from __future__ import annotations

from typing import Optional, Union

from pydantic.v1 import Field

from oold.model.v1 import LinkedBaseModel


class Subschema1(LinkedBaseModel):
    class Config:
        schema_extra = {"title": "Subschema1"}

    subprop0: Optional[str] = Field(None, custom_key="custom_value_0")
    subprop1: Optional[str] = Field(None, custom_key="custom_value_1")


class Subschema2(LinkedBaseModel):
    class Config:
        schema_extra = {"title": "Subschema2"}

    subprop0: Optional[str] = Field(None, custom_key="custom_value_0")
    subprop2: Optional[str] = Field(None, custom_key="custom_value_2")


class Example(LinkedBaseModel):
    class Config:
        schema_extra = {"title": "Example"}

    type: Optional[str] = ["example"]
    prop1: Optional[str] = Field(None, custom_key="custom_value")
    prop2: Optional[Union[Subschema1, Subschema2]] = Field(
        None, custom_key="custom_value"
    )



import ast
import inspect
import json
from enum import Enum
from functools import partial
from operator import is_
from typing import Any, Callable, Dict, Optional, Tuple, TypeVar, Union

import jsondiff
import pydantic_core.core_schema as core_schema
import pyld
import yaml
from pydantic import BaseModel, create_model
from pydantic.json_schema import GenerateJsonSchema, JsonSchemaMode, JsonSchemaValue
from pydantic.v1 import BaseModel as BaseModel_v1
from pydantic.v1 import create_model as create_model_v1
from pydantic_core import CoreSchema, to_jsonable_python
from pyld import jsonld
from typing_extensions import Literal

E = TypeVar("E", bound=Enum)


# from https://stackoverflow.com/a/79229811
# see also https://stackoverflow.com/a/78943193
def enum_docstrings(enum: type[E]) -> type[E]:
    '''Attach docstrings to enum members

    Docstrings are string literals that appear directly below the enum member
    assignment expression:

    ```
    @enum_docstrings
    class SomeEnum(Enum):
        """Docstring for the SomeEnum enum"""

        foo_member = "foo_value"
        """Docstring for the foo_member enum member"""

    SomeEnum.foo_member.__doc__  # 'Docstring for the foo_member enum member'
    ```

    '''
    try:
        mod = ast.parse(inspect.getsource(enum))
    except OSError:
        # no source code available
        return enum

    if mod.body and isinstance(class_def := mod.body[0], ast.ClassDef):
        # An enum member docstring is unassigned if it is the exact same object
        # as enum.__doc__.
        unassigned = partial(is_, enum.__doc__)
        names = enum.__members__.keys()
        member: E | None = None
        for node in class_def.body:
            # case ast.Assign(targets=[ast.Name(id=name)]) if name in names:
            if isinstance(node, ast.Assign):
                if len(node.targets) == 1 and isinstance(node.targets[0], ast.Name):
                    name = node.targets[0].id
                    if name in names:
                        # Enum member assignment, look for a docstring next
                        member = enum[name]
                        continue

            # case ast.Expr(value=ast.Constant(value=str(docstring)))
            # if member and unassigned(member.__doc__):
            if (
                isinstance(node, ast.Expr)
                and isinstance(node.value, ast.Constant)
                and isinstance(node.value.value, str)
                and member
                and unassigned(member.__doc__)
            ):
                # docstring immediately following a member assignment
                docstring = node.value.value
                member.__doc__ = docstring
            else:
                pass

            member = None

    return enum


class OOLDJsonSchemaGenerator(GenerateJsonSchema):
    def generate(
        self, schema: CoreSchema, mode: JsonSchemaMode = "validation"
    ) -> JsonSchemaValue:
        return super().generate(schema, mode)

    def nullable_schema(self, schema: Dict[str, Any]) -> Dict[str, Any]:
        """Override to handle nullable schemas - do not add 'null' type
        since optional fields are already handled by 'required'."""
        inner_json_schema = self.generate_inner(schema["schema"])
        return inner_json_schema

    def enum_schema(self, schema: core_schema.EnumSchema) -> JsonSchemaValue:
        """Generates a JSON schema that matches an Enum value.

        Args:
            schema: The core schema.

        Returns:
            The generated JSON schema.
        """
        enum_type = enum_docstrings(schema["cls"])
        description = (
            None if not enum_type.__doc__ else inspect.cleandoc(enum_type.__doc__)
        )
        if (
            description == "An enumeration."
        ):  # This is the default value provided by enum.EnumMeta.__new__; don't use it
            description = None
        result: dict[str, Any] = {
            "title": enum_type.__name__,
            "description": description,
        }
        result = {k: v for k, v in result.items() if v is not None}

        expected = [to_jsonable_python(v.value) for v in schema["members"]]

        result["enum"] = expected

        # interate over all enum values and collect their docstrings
        enum_descriptions = []
        for v in enum_type:
            if v.value in expected and v.__doc__:
                enum_descriptions.append(f"{inspect.cleandoc(v.__doc__)}")
            else:
                enum_descriptions.append(f"{v.value}")
        if "options" not in result:
            result["options"] = {}
        result["options"]["enum_titles"] = enum_descriptions

        result["x-enum-varnames"] = [v.name for v in enum_type if v.value in expected]
        result["x-enum-descriptions"] = enum_descriptions

        types = {type(e) for e in expected}
        if isinstance(enum_type, str) or types == {str}:
            result["type"] = "string"
        elif isinstance(enum_type, int) or types == {int}:
            result["type"] = "integer"
        elif isinstance(enum_type, float) or types == {float}:
            result["type"] = "number"
        elif types == {bool}:
            result["type"] = "boolean"
        elif types == {list}:
            result["type"] = "array"

        return result


class SchemaExportMode(str, Enum):
    """Enum for schema export modes."""

    FULL = "full"
    """Export the full schema including all base classes.
    Equivalent to cutoff at the BaseModel class"""
    PARTIAL = "partial"
    """Export the schema of a model up to the specified cutoff
    base class, default the direct parent class"""


class PartialSchemaExportMode(str, Enum):
    """Enum for partial schema export modes."""

    BASE_CLASS_CUTOFF = "base_class_cutoff"
    """Export the schema of a model up to the specified base class
    by cutoff the class hierarchy"""
    BASE_CLASS_DIFF = "base_class_diff"
    """Export the schema of a model up to the specified base class
    by calculating the diff to the base class schema."""


class GenericLinkedBaseModel:
    def _object_to_iri(self, d, exclude_none=False):
        for name in list(d.keys()):  # force copy of keys for inline-delete
            if name in self.__iris__:
                d[name] = self.__iris__[name]
            if exclude_none and d[name] is None:
                del d[name]
        return d

    @staticmethod
    def remove_none(d: Dict) -> Dict:
        """Remove None values from a dictionary recursively."""
        if isinstance(d, dict):
            return {
                k: GenericLinkedBaseModel.remove_none(v)
                for k, v in d.items()
                if v is not None
            }
        elif isinstance(d, list):
            return [GenericLinkedBaseModel.remove_none(i) for i in d]
        else:
            return d

    @classmethod
    def export_schema(
        cls,
        mode: Optional[SchemaExportMode] = SchemaExportMode.FULL,
        cutoff_base_cls: Optional[
            Union[Union[BaseModel, BaseModel_v1], Tuple[Union[BaseModel, BaseModel_v1]]]
        ] = None,
        partial_mode: Optional[
            PartialSchemaExportMode
        ] = PartialSchemaExportMode.BASE_CLASS_CUTOFF,
        serialize: Optional[Literal["json", "yaml"]] = None,
    ) -> Dict:
        """Export the schema of the model as a dictionary."""
        schema = export_schema(cls, mode, cutoff_base_cls, partial_mode)
        if serialize == "json":
            return json.dumps(schema, indent=2)
        elif serialize == "yaml":
            _ignore_aliases = yaml.Dumper.ignore_aliases
            yaml.Dumper.ignore_aliases = lambda *args: True
            yaml_doc = yaml.dump(schema, indent=2)
            yaml.Dumper.ignore_aliases = _ignore_aliases
            return yaml_doc
        return schema


def get_jsonld_context_loader(model_cls, model_type) -> Callable:
    """to overwrite the default jsonld document loader to load
    relative context from the osl"""

    classes = [model_cls]
    i = 0
    while 1:
        try:
            cls = classes[i]
            if cls == model_type:
                break
        except IndexError:
            break
        i += 1
        classes[i:i] = [base for base in cls.__bases__ if base not in classes]

    schemas = {}
    for base_class in classes:
        schema = {}
        if model_type == BaseModel:
            if hasattr(base_class, "model_config"):
                schema = base_class.model_config.get("json_schema_extra", {})
        if model_type == BaseModel_v1:
            if hasattr(base_class, "__config__"):
                schema = base_class.__config__.schema_extra
        iri = schema.get("iri", None)
        title = schema.get("title", None)
        if iri is None and title is None:
            continue
        if iri is None:
            schema.get("title")
            if title in ["Entity", "Category", "Property"]:
                iri = "Category:" + schema.get("title")
            else:
                iri = "Category:" + "OSW" + schema.get("uuid").replace("-", "")
        # iri = base_class.get_iri() # does not work on class level
        schemas[iri] = schema

    # print(schemas)

    def loader(url, options=None):
        if options is None:
            options = {}
        # print("Requesting", url)
        if "/wiki/" in url:
            url = url.split("/")[-1].split("?")[0]
        if url in schemas:
            schema = schemas[url]

            doc = {
                "contentType": "application/json",
                "contextUrl": None,
                "documentUrl": url,
                "document": schema,
            }
            # print("Loaded", doc)
            return doc

        else:
            requests_loader = pyld.documentloader.requests.requests_document_loader()
            return requests_loader(url, options)

    return loader


def export_jsonld(model_instance, model_type) -> Dict:
    """Return the RDF representation of the object as JSON-LD."""

    # serialize the model to a dictionary
    # to_string().to_json() roundtrips is needed to serialize enums correctly
    if model_type == BaseModel:
        # get the context from self.ConfigDict.json_schema_extra["@context"]
        context = model_instance.model_config.get("json_schema_extra", {}).get(
            "@context", {}
        )
        data = json.loads(model_instance.model_dump_json(exclude_none=True))
    if model_type == BaseModel_v1:
        context = model_instance.__class__.__config__.schema_extra.get("@context", {})
        data = json.loads(model_instance.json(exclude_none=True))

    if "id" not in data and "@id" not in data:
        data["id"] = model_instance.get_iri()
    jsonld_dict = {"@context": context, **data}
    jsonld.set_document_loader(
        get_jsonld_context_loader(model_instance.__class__, model_type)
    )
    jsonld_dict = jsonld.expand(jsonld_dict)
    if isinstance(jsonld_dict, list):
        jsonld_dict = jsonld_dict[0]
    return jsonld_dict


def import_jsonld(model_type, jsonld_dict: Dict, _types: Dict[str, type]):
    """Return the object instance from the JSON-LD representation."""
    # ToDo: apply jsonld frame with @id restriction
    # get the @type from the jsonld_dict
    type_iri = jsonld_dict.get("@type", None)
    # if type_iri is None, return None
    if type_iri is None:
        return None
    # if type_iri is a list, get the first element
    if isinstance(type_iri, list):
        type_iri = type_iri[0]
    # get the class from the _types dict
    # Todo: IRI normalization
    type_iri = type_iri.split("/")[-1]
    model_cls = _types.get(type_iri, None)
    # if model_type is None, return None
    if model_cls is None:
        return None
    if model_type == BaseModel:
        # get the context from self.ConfigDict.json_schema_extra["@context"]
        context = model_cls.model_config.get("json_schema_extra", {}).get(
            "@context", {}
        )
    if model_type == BaseModel_v1:
        context = model_cls.__config__.schema_extra.get("@context", {})
    jsonld.set_document_loader(get_jsonld_context_loader(model_cls, model_type))
    jsonld_dict = jsonld.compact(jsonld_dict, context)
    if "@context" in jsonld_dict:
        del jsonld_dict["@context"]
    return model_cls(**jsonld_dict)


def _get_schema(model_cls):
    """Return the schema of the model as a dictionary."""
    if issubclass(model_cls, BaseModel):
        return model_cls.model_json_schema(
            ref_template="#/$defs/{model}",
            schema_generator=OOLDJsonSchemaGenerator,
        )
    elif issubclass(model_cls, BaseModel_v1):
        return model_cls.schema(ref_template="#/$defs/{model}")


def _inverse_preprocess(schema: Dict):
    """Inverse preprocess the JSON schemas to remove
    the $refs generated from x-oold-range annotations."""

    def handle_property(property):
        if "range" in property:
            # restore a string type, remove allOf and $ref
            # if they (or one element) match the range value
            if isinstance(property["range"], str):
                if "allOf" in property:
                    # remove the array element that matches the range
                    # or a self-referencing $ref
                    property["allOf"] = [
                        item
                        for item in property["allOf"]
                        if item["$ref"] in ["#", property["range"]]
                    ]
                if "$ref" in property:
                    if property["$ref"] in ["#", property["range"]]:
                        del property["$ref"]
            else:
                if "allOf" in property:
                    # remove the array element that matches the range
                    property["allOf"] = [
                        item for item in property["allOf"] if item == property["range"]
                    ]
            if "allOf" in property and len(property["allOf"]) == 0:
                del property["allOf"]
            if "type" not in property:
                property["type"] = "string"
        if "properties" in property:
            _inverse_preprocess(property)

    for property_key in schema.get("properties", {}):
        property = schema["properties"][property_key]
        if "x-oold-required-iri" in property:
            # remove the x-oold-required-iri property
            # add property to required
            del property["x-oold-required-iri"]
            if "required" not in schema:
                schema["required"] = []
            if property_key not in schema["required"]:
                schema["required"].append(property_key)

        if "items" in property:
            if "range" in property["items"]:
                property["items"]["range"] = property["range"]
                del property["range"]
            handle_property(property["items"])

        else:
            handle_property(property)
    return schema


def _export_schema_from_dynamic_model(
    model_cls: Union[BaseModel, BaseModel_v1]
) -> Dict:
    """Export the OO-LD schema of a single pydantic model.
    Class hierarchy is not considered, only the model itself
    by generating a model copy without base classes.
    """

    if issubclass(model_cls, BaseModel):
        field_dict = {}
        for field_name, field in model_cls.model_fields.items():
            field_dict[field_name] = (field.annotation, field)

        # create model dynamically
        model_cls_copy = create_model(
            model_cls.__name__,
            __config__=model_cls.model_config,
            __doc__=model_cls.__doc__,
            __module__=model_cls.__module__,
            **field_dict,
            # __base__=BaseModel,
        )

    elif issubclass(model_cls, BaseModel_v1):
        model_cls: BaseModel_v1 = model_cls  # type: ignore
        field_dict = {}
        for field_name, model_field in model_cls.__fields__.items():
            field_dict[field_name] = (model_field.annotation, model_field.field_info)

        # create model dynamically
        model_cls_copy = create_model_v1(
            model_cls.__name__,
            __config__=model_cls.__config__,
            __doc__=model_cls.__doc__,
            __module__=model_cls.__module__,
            **field_dict,
            # __base__=BaseModel_v1,
        )

    my_schema_full = _get_schema(model_cls_copy)

    return my_schema_full


def export_schema(
    model_cls: Union[BaseModel, BaseModel_v1],
    mode: Optional[SchemaExportMode] = SchemaExportMode.FULL,
    cutoff_base_cls: Optional[
        Union[Union[BaseModel, BaseModel_v1], Tuple[Union[BaseModel, BaseModel_v1]]]
    ] = None,
    partial_mode: Optional[
        PartialSchemaExportMode
    ] = PartialSchemaExportMode.BASE_CLASS_CUTOFF,
) -> Dict:
    """Export the OO-LD schema of the model as a JSON-SCHEMA with JSON-LD context"""

    if mode == SchemaExportMode.FULL:
        # export the full schema including all base classes
        result_schema = _get_schema(model_cls)
        if result_schema is None:
            raise ValueError(f"Model {model_cls.__name__} has no schema.")

    elif mode == SchemaExportMode.PARTIAL:
        # export the schema of a model up to the specified cutoff base class

        baseclass_schema = {}
        imports = []

        if cutoff_base_cls is None:
            cutoff_base_cls = model_cls.__bases__
        if not isinstance(cutoff_base_cls, tuple):
            cutoff_base_cls = (cutoff_base_cls,)

        # if partial_mode == PartialSchemaExportMode.BASE_CLASS_CUTOFF:

        for baseclass in cutoff_base_cls:
            if baseclass in [BaseModel, BaseModel_v1]:
                continue
            schema = _get_schema(baseclass)
            if schema is not None:
                # ToDO: Use deepmerge
                baseclass_schema = {**baseclass_schema, **schema}

            # ToDo: determine import references
            # baseclass_oswid = ""
            # if baseclass.__name__ in toplevel_classes:
            #     baseclass_oswid = toplevel_classes[baseclass.__name__]
            # else:
            #     baseclass_oswid = get_osw_id(
            #         baseclass.__config__.schema_extra["uuid"]
            #     )

            # imports.append(f"/wiki/Category:{baseclass_oswid}?action=raw&slot=jsonschema")
            # try the follwing schema extra attributes: $id, iri, class name
            import_ref = None
            if issubclass(model_cls, BaseModel):
                import_ref = baseclass.model_config.get("json_schema_extra", {}).get(
                    "$id", None
                )
                if import_ref is None:
                    import_ref = baseclass.model_config.get(
                        "json_schema_extra", {}
                    ).get("iri", None)
                if import_ref is None:
                    import_ref = baseclass.__name__

            if issubclass(model_cls, BaseModel_v1):
                import_ref = baseclass.__config__.schema_extra.get("$id", None)
                if import_ref is None:
                    import_ref = baseclass.__config__.schema_extra.get("iri", None)
                if import_ref is None:
                    import_ref = baseclass.__name__

            if import_ref is not None:
                imports.append(import_ref)

        if partial_mode == PartialSchemaExportMode.BASE_CLASS_DIFF:
            # option 1: calculate a schema diff
            # implementation not yet complete

            for baseclass in cutoff_base_cls:
                if baseclass in [BaseModel, BaseModel_v1]:
                    continue
                schema = _get_schema(baseclass)
                if schema is not None:
                    # ToDO: Use deepmerge
                    baseclass_schema = {**baseclass_schema, **schema}

            model_schema_full = _get_schema(model_cls)
            model_schema_diff = jsondiff.diff(
                baseclass_schema, model_schema_full, marshal=True
            )

            # implementation note: jsonpath does not work
            # since we cannot reconstruct the diff document
            # import jsonpatch
            # patch = jsonpatch.make_patch(baseclass_schema, model_schema_full)
            # print("Patch: ", json.dumps(patch.patch, indent=2))
            # this fails due to missing root paths
            # see https://github.com/stefankoegl/python-json-patch/issues/153
            # print("Diff: ", json.dumps(patch.apply({}), indent=2))

            required = []
            if "required" in model_schema_diff:
                if "$insert" in model_schema_diff["required"]:
                    for r in model_schema_diff["required"]["$insert"]:
                        required.append(r[1])
                    del model_schema_diff["required"]
            model_schema_diff["required"] = required
            if "$delete" in model_schema_diff:
                del model_schema_diff["$delete"]
            # todo: handle all "$insert", "$delete" and "$replace" keys

        elif partial_mode == PartialSchemaExportMode.BASE_CLASS_CUTOFF:
            # option 2: export the schema of a model up to the specified
            # base class by cutoff the class hierarchy
            model_schema_diff = _export_schema_from_dynamic_model(model_cls)

        context = None
        if "@context" in model_schema_diff:
            context = model_schema_diff.pop("@context")

            if not isinstance(context, list):
                context = [context]
            missing_imports = []
            for imp in imports:
                if imp not in context:
                    missing_imports.append(imp)
            context = [*missing_imports, *context]

        defs = None
        if "definitions" in model_schema_diff:
            defs = model_schema_diff.pop("definitions")
        if "$defs" in model_schema_diff:
            if defs is None:
                defs = {}
            defs = {**defs, **model_schema_diff.pop("$defs")}
        result_schema = {
            "@context": context,
            "$defs": defs,
            **json.loads(
                json.dumps(model_schema_diff)
                .replace("title_", "title*")
                .replace("description_", "description*")
                .replace("$$ref", "$ref")
            ),
        }

    # ToDo: move this to general json utils

    # if schema has a $ref on the root level, resolve it by following the reference
    # e.g. {"$defs": {"Entity": {"title": "Entity"}}, "$ref": "#/$defs/Entity"}
    # => {"title": "Entity"}
    if "$ref" in result_schema:
        ref = result_schema["$ref"]
        if ref.startswith("#/"):
            # follow the ref path elements
            ref_path = ref.split("/")[1:]  # remove the leading "#/"
            ref_target = result_schema
            for path_element in ref_path:
                if path_element in ref_target:
                    ref_target = ref_target[path_element]
                else:
                    raise ValueError(f"Reference {ref} not found in schema.")
            result_schema = {**ref_target, **result_schema}

            # rewrite other $ref to the current schema
            # => e.g. replace all {"$ref": "#/$defs/Entity"} with {"$ref": "#"}
            # iterate over all result_schema['properties']
            # and schema['properties']['items'] and remove the $ref
            refs = []
            for property_key in result_schema.get("properties", {}):
                property = result_schema["properties"][property_key]
                if "$ref" in property:
                    refs.append(property["$ref"])
                    if property["$ref"] == ref:
                        property["$ref"] = "#"
                if "items" in property:
                    if "$ref" in property["items"]:
                        refs.append(property["items"]["$ref"])
                        if property["items"]["$ref"] == ref:
                            property["items"]["$ref"] = "#"
            # if non of the refs is a subpath of the ref, remove $defs element
            # in case it is at root level of $defs
            if not any(_ref.startswith(ref + "/") for _ref in refs):
                if ref.split("/")[:-1] == ["#", "$defs"]:
                    del result_schema["$defs"][ref.split("/")[-1]]

            del result_schema["$ref"]
    result_schema = _inverse_preprocess(result_schema)
    return result_schema



from typing import Dict, Optional, Type, Union

from nicegui.element import Element
from nicegui.events import GenericEventArguments, Handler, handle_event
from typing_extensions import Self

from oold.model import LinkedBaseModel
from oold.model.v1 import LinkedBaseModel as LinkedBaseModel_v1


# class JsonEditor(ui.element,
class JsonEditor(
    Element,
    component="../vue/src/jsoneditor.vue",
    dependencies=["../vue/node_modules/@json-editor/json-editor/dist/jsoneditor.js"],
):
    def __init__(
        self,
        options: Optional[Dict] = None,
        on_change: Optional[Handler[GenericEventArguments]] = None,
    ) -> None:
        """JsonEditor

        An element that integrates the `JSON-SCHEMA form generator library
        <https://github.com/json-editor/json-editor>`.
        """
        super().__init__()
        self._props["options"] = options or {
            # "theme": 'tailwind',
            "theme": "bootstrap4",
            # "iconlib": 'spectre',
            "schema": {
                "required": ["test"],
                "properties": {"test": {"type": "string"}},
            },
            # "startval": {}
        }

        self._change_handlers = [on_change] if on_change else []

        self.on("change", self.handle_change)

    def handle_change(self, e: GenericEventArguments) -> None:
        for handler in self._change_handlers:
            handle_event(handler, e)

    def on_change(self, callback: Handler[GenericEventArguments]) -> Self:
        """Add a callback to be invoked when the user touches the joystick."""
        self._change_handlers.append(callback)
        return self

    # def clear(self):
    #    """Clear the signature."""
    #    self.run_method('clear')


class OswEditor(JsonEditor):
    def __init__(
        self,
        options: Optional[Dict] = None,
        on_change: Optional[Handler[GenericEventArguments]] = None,
        entity: Union[Type[LinkedBaseModel], Type[LinkedBaseModel_v1]] = None,
    ) -> None:
        options = options or {
            # "theme": 'tailwind',
            "theme": "bootstrap4",
            # "iconlib": 'spectre',
            "schema": {
                "required": ["test"],
                "properties": {"test": {"type": "string"}},
            },
            # "startval": {}
        }
        if entity is not None:
            options["schema"] = entity.export_schema()
        self.entity = entity
        super().__init__(options, on_change)

    def handle_change(self, e: GenericEventArguments) -> None:
        print(e)
        try:
            self.entity(**e.args)
        except Exception as ex:
            print("No valid instance: ", ex)
        super().handle_change(e)



import json
import os
import re
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Dict, List, Optional

import datamodel_code_generator
from datamodel_code_generator import DataModelType, InputFileType, generate
from pydantic import BaseModel

from oold.utils.codegen import OOLDJsonSchemaParserFixedRefs as OOLDJsonSchemaParser


class Generator:
    class GenerateParams(BaseModel):
        json_schemas: List[Dict]
        """JSON SCHEMA source(s)"""
        preprocess: bool = True
        """Preprocess the JSON schemas before generating the models"""
        main_schema: Optional[str] = None
        """File name of the main schema"""
        output_model_type: Optional[DataModelType] = (
            DataModelType.PydanticV2BaseModel,
        )
        """Output model type, e.g. PydanticV2BaseModel or PydanticBaseModel"""
        output_model_path: Optional[Path] = (
            Path(__file__).parent / "model" / "example.py"
        )
        """Output model path, if not set the model will be generated
        in the current directory"""
        working_dir_path: Optional[Path] = None
        """Working directory to store intermedia files
        and the generated partial models"""
        generate_init_py_files: bool = True
        """Generate __init__.py files along the output_model_path"""

    def generate(
        self,
        params: GenerateParams,
    ):
        if params.preprocess:
            self.preprocess(
                Generator.PreprocessParams(json_schemas=params.json_schemas)
            )

        # monkey patch class
        datamodel_code_generator.parser.jsonschema.JsonSchemaParser = (
            OOLDJsonSchemaParser
        )

        with TemporaryDirectory() as temporary_directory_name:
            temporary_directory = Path(temporary_directory_name)
            if params.working_dir_path is not None:
                temporary_directory = params.working_dir_path

            input = Path(temporary_directory)
            if params.main_schema is not None:
                input = Path(temporary_directory / Path(params.main_schema))

            output = params.output_model_path
            if params.generate_init_py_files:
                # generate __init__.py files in every subdirectory
                # of the output model path that does not exist yet
                # output may be a file or a directory

                # check if output is a file or a directory
                target_dir = output
                if params.main_schema is not None:
                    target_dir = output.parent

                # interate over the target_dir path, starting at the top level dir
                # e.g. 'C:' or '/var'
                for segment in target_dir.parts:
                    # create the segment path
                    segment_path = Path(
                        *target_dir.parts[: target_dir.parts.index(segment) + 1]
                    )
                    # check if the segment path exists
                    if not segment_path.exists():
                        # create the __init__.py file
                        os.makedirs(segment_path, exist_ok=False)
                        init_file = segment_path / "__init__.py"
                        with open(init_file, "w", encoding="utf-8") as f:
                            f.write("# Generated by oold.generator\n")

            for schema in params.json_schemas:
                name = schema["id"]
                os.makedirs(
                    os.path.dirname(Path(temporary_directory / (name + ".json"))),
                    exist_ok=True,
                )
                with open(
                    Path(temporary_directory / (name + ".json")), "w", encoding="utf-8"
                ) as f:
                    schema_str = json.dumps(
                        schema, ensure_ascii=False, indent=4
                    ).replace("dollarref", "$ref")
                    # print(schema_str)
                    f.write(schema_str)

            if params.output_model_type == DataModelType.PydanticV2BaseModel:
                base_class = "oold.model.LinkedBaseModel"
            else:
                base_class = "oold.model.v1.LinkedBaseModel"
            generate(
                input_=input,
                # json_schema,
                input_file_type=InputFileType.JsonSchema,
                # input_filename="Foo.json",
                output=output,
                # set up the output model types
                output_model_type=params.output_model_type,
                # custom_template_dir=Path(model_dir_path),
                field_include_all_keys=True,
                base_class=base_class,
                # use_default = True,
                enum_field_as_literal="all",
                use_title_as_name=True,
                use_schema_description=True,
                use_field_description=True,
                encoding="utf-8",
                use_double_quotes=True,
                collapse_root_models=True,
                reuse_model=True,
                # create MyEnum(str, Enum) instead of MyEnum(Enum)
                use_subclass_enum=True,
                additional_imports=["pydantic.ConfigDict"]
                if params.output_model_type == DataModelType.PydanticV2BaseModel
                else [],
                apply_default_values_for_required_fields=True,
            )

            if params.main_schema is not None:
                content = ""
                with open(output, "r", encoding="utf-8") as f:
                    content = f.read()
                os.remove(output)

                content = re.sub(
                    r"(UUID = Field\(...)",
                    r"UUID = Field(default_factory=uuid4",
                    content,
                )  # enable default value for uuid

                if params.output_model_type == DataModelType.PydanticBaseModel:
                    # we are now using pydantic.v1
                    # pydantic imports lead to uninitialized fields
                    # (FieldInfo still present)
                    content = re.sub(
                        r"(from pydantic import)", "from pydantic.v1 import", content
                    )

                # write the content to the file
                with open(output, "w", encoding="utf-8") as f:
                    f.write(content)

    class PreprocessParams(BaseModel):
        json_schemas: List[Dict]
        """JSON SCHEMA source(s)"""

    def preprocess(self, params: PreprocessParams):
        for schema in params.json_schemas:
            # schema = self.merge_property_schemas(schema)
            for property_key in schema.get("properties", {}):
                property = schema["properties"][property_key]
                if "range" in property:
                    if "type" in property:
                        del property["type"]
                    # if range is a string we create a allOf with a ref to the range
                    if isinstance(property["range"], str):
                        property["allOf"] = [{"$ref": property["range"]}]
                    else:
                        property["$ref"] = property["range"]
                    if "required" in schema and property_key in schema["required"]:
                        # if no default value is set, remove the property from required
                        if "default" not in property:
                            schema["required"].remove(property_key)
                        if "x-oold-required-iri" not in property:
                            property["x-oold-required-iri"] = True
                if "items" in property:
                    if "range" in property["items"]:
                        if "type" in property["items"]:
                            del property["items"]["type"]
                        if isinstance(property["items"]["range"], str):
                            property["items"]["allOf"] = [
                                {"$ref": property["items"]["range"]}
                            ]
                        else:
                            property["items"]["$ref"] = property["items"]["range"]
                        property["range"] = property["items"]["range"]
                        if "required" in schema and property_key in schema["required"]:
                            # if no default value is set,
                            # remove the property from required
                            if "default" not in property["items"]:
                                schema["required"].remove(property_key)
                            if "x-oold-required-iri" not in property:
                                property["x-oold-required-iri"] = True

                    if "properties" in property["items"]:
                        self.preprocess(
                            Generator.PreprocessParams(json_schemas=[property["items"]])
                        )

                if "properties" in property:
                    self.preprocess(Generator.PreprocessParams(json_schemas=[property]))



from pathlib import Path
from typing import Any, Dict

from datamodel_code_generator import load_yaml_from_path
from datamodel_code_generator.model import pydantic as pydantic_v1_model
from datamodel_code_generator.model import pydantic_v2 as pydantic_v2_model
from datamodel_code_generator.parser.jsonschema import (
    DataType,
    JsonSchemaObject,
    JsonSchemaParser,
    get_special_path,
)

# https://docs.pydantic.dev/1.10/usage/schema/#schema-customization
# https://docs.pydantic.dev/latest/concepts/json_schema/#using-json_schema_extra-with-a-dict
# https://docs.pydantic.dev/latest/concepts/json_schema/#field-level-customization


class PydanticV1Config(pydantic_v1_model.Config):
    # schema_extra: Optional[Dict[str, Any]] = None
    schema_extra: str = None


class PydanticV2Config(pydantic_v2_model.ConfigDict):
    # schema_extra: Optional[Dict[str, Any]] = None
    json_schema_extra: str = None


class OOLDJsonSchemaParser(JsonSchemaParser):
    """Custom parser for OO-LD schemas.
    You can use this class directly or monkey-patch the datamodel_code_generator module:
    `datamodel_code_generator.parser.jsonschema.JsonSchemaParser = OOLDJsonSchemaParser`
    """

    def set_additional_properties(self, name: str, obj: JsonSchemaObject) -> None:
        schema_extras = repr(obj.extras)  # keeps 'False' and 'True' boolean literals
        if self.data_model_type == pydantic_v1_model.BaseModel:
            self.extra_template_data[name]["config"] = PydanticV1Config(
                schema_extra=schema_extras
            )
        if self.data_model_type == pydantic_v2_model.BaseModel:
            self.extra_template_data[name]["config"] = PydanticV2Config(
                json_schema_extra=schema_extras
            )
        return super().set_additional_properties(name, obj)

    # fixes https://github.com/koxudaxi/datamodel-code-generator/issues/2403
    def parse_combined_schema(
        self,
        name: str,
        obj: JsonSchemaObject,
        path: list[str],
        target_attribute_name: str,
    ) -> list[DataType]:
        base_object = obj.dict(
            exclude={target_attribute_name}, exclude_unset=True, by_alias=True
        )
        # base_object["extras"] = obj.extras # rename to
        # '#-datamodel-code-generator-#-extras-#-special-#' by alias export
        # if "subprop0" in base_object["properties"]:
        #     base_object["properties"]["subprop0"]["extras"] = obj.extras
        combined_schemas: list[JsonSchemaObject] = []
        refs = []
        for index, target_attribute in enumerate(
            getattr(obj, target_attribute_name, [])
        ):
            if target_attribute.ref:
                combined_schemas.append(target_attribute)
                refs.append(index)
                # TODO: support partial ref
            else:
                # combined_schemas.append(
                #     self.SCHEMA_OBJECT_TYPE.parse_obj(
                #         self._deep_merge(
                #             base_object,
                #             target_attribute.dict(exclude_unset=True, by_alias=True)
                #         )
                #     )
                # )
                so = self.SCHEMA_OBJECT_TYPE.parse_obj(
                    self._deep_merge(
                        base_object,
                        target_attribute.dict(exclude_unset=True, by_alias=True),
                    )
                )
                if hasattr(so, "properties") and hasattr(obj, "properties"):
                    if so.properties is not None and obj.properties is not None:
                        for k, v in so.properties.items():
                            if k in obj.properties:
                                if obj.properties[k].extras:
                                    v.extras = self._deep_merge(
                                        v.extras, obj.properties[k].extras
                                    )
                            x_of_properties = getattr(
                                target_attribute, "properties", {}
                            )
                            if k in x_of_properties:
                                if x_of_properties[k].extras:
                                    v.extras = self._deep_merge(
                                        v.extras, x_of_properties[k].extras
                                    )
                combined_schemas.append(so)

        parsed_schemas = self.parse_list_item(
            name,
            combined_schemas,
            path,
            obj,
            singular_name=False,
        )
        common_path_keyword = f"{target_attribute_name}Common"
        return [
            self._parse_object_common_part(
                name,
                obj,
                [*get_special_path(common_path_keyword, path), str(i)],
                ignore_duplicate_model=True,
                fields=[],
                base_classes=[d.reference],
                required=[],
            )
            if i in refs and d.reference
            else d
            for i, d in enumerate(parsed_schemas)
        ]

    def parse_enum(
        self,
        name: str,
        obj: JsonSchemaObject,
        path: list[str],
        singular_name: bool = False,  # noqa: FBT001, FBT002
        unique: bool = True,  # noqa: FBT001, FBT002
    ) -> DataType:
        """Override to handle enum schemas - add enum descriptions as docstrings."""
        schema = super().parse_enum(name, obj, path, singular_name, unique)
        ref = schema.reference
        for r in self.results:
            if r.reference == ref:
                if "options" in obj.extras and "enum_titles" in obj.extras["options"]:
                    for i, v in enumerate(obj.extras["options"]["enum_titles"]):
                        if "description" not in r.fields[i].extras:
                            r.fields[i].extras["description"] = v
        return schema


class OOLDJsonSchemaParserFixedRefs(OOLDJsonSchemaParser):
    """Overwrite # overwrite the original `_get_ref_body_from_remote` function to fix
    wrongly composed paths. This issue occurs only when using this parser class directy
    and occurs not if used through mokey patching and
    `datamodel_code_generator.generate()`.
    Only relevant for schema refs in subdirs
    """

    def _get_ref_body_from_remote(self, resolved_ref: str) -> Dict[Any, Any]:
        # default behaviour:  full_path = self.base_path / resolved_ref
        # fix: merge the paths correctly:
        # resolved_ref:  'C:/Users/Example/Git/OO-LD/oold-python/src/oold/example/src/Bar.json' # noqa: E501
        # self.base_path 'C:/Users/Example/Git/OO-LD/oold-python/src/oold/example/src'
        # => full_path:  'C:/Users/Example/Git/OO-LD/oold-python/src/oold/example/src/Bar.json' # noqa: E501
        # resolved_ref:  'C:/Users/Example/Git/OO-LD/oold-python/src/oold/example/src/Users/Example/Git/OO-LD/oold-python/src/oold/example/src/bar2/Bar2.json' # noqa: E501
        # self.base_path 'C:/Users/Example/Git/OO-LD/oold-python/src/oold/example/src'
        # => full_path:  'C:/Users/Example/Git/OO-LD/oold-python/src/oold/example/src/bar2/Bar2.json' # noqa: E501

        # get path without drive and scheme, so that it works on Windows and Linux
        resolved_ref = Path(resolved_ref).as_posix()  # remove drive and scheme
        resolved_ref = resolved_ref.split(self.base_path.as_posix().split(":", 1)[-1])[
            -1
        ].lstrip(
            "/"
        )  # remove base path, leading '/'
        full_path = self.base_path / Path(resolved_ref)
        return self.remote_object_cache.get_or_put(
            str(full_path),
            default_factory=lambda _: load_yaml_from_path(full_path, self.encoding),
        )



import json

import pytest
import rdflib
from jsondiff import diff
from pyld import jsonld

from oold.utils.transform import json_to_json, jsonld_to_jsonld


def test_simple_json():
    input_data = {"type": "Human", "label": "Jane Doe"}

    input_context = {
        "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
        "label": "rdfs:label",
        "type": "@type",
        "ex": "https://another-example.org/",
        "Human": "ex:Human",
    }
    mapping_context = {
        "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
        "schema": "https://schema.org/",
        "name*": "rdfs:label",
        "name": "schema:name",
        "type": "@type",
        "ex": "https://another-example.org/",
        "Person*": "ex:Human",
        "Person": "schema:Person",
    }
    output_data = json_to_json(input_data, mapping_context, input_context)
    expected_output_data = {
        "type": "Person",
        "name": "Jane Doe",
    }

    print("Output Data:", output_data)
    assert (
        output_data == expected_output_data
    ), f"Expected {expected_output_data}, but got {output_data}"


def test_complex_graph():
    graph = {
        "@context": {
            "schema": "http://schema.org/",
            "demo": "https://oo-ld.github.io/demo/",
            "name": "schema:name",
            "full_name": "demo:full_name",
            "label": "demo:label",
            "works_for": {"@id": "schema:worksFor", "@type": "@id"},
            "is_employed_by": {"@id": "demo:is_employed_by", "@type": "@id"},
            "employes": {"@id": "schema:employes", "@type": "@id"},
            "type": "@type",
            "id": "@id",
        },
        "@graph": [
            {
                "id": "demo:person1",
                "type": "schema:Person",
                "name": "Person1",
                "works_for": "demo:organizationA",
            },
            {
                "id": "demo:person2",
                "type": "schema:Person",
                "full_name": "Person2",
                "is_employed_by": "demo:organizationA",
            },
            {"id": "demo:person3", "type": "schema:Person", "name": "Person3"},
            {
                "id": "demo:organizationA",
                "type": "schema:Organization",
                "label": "organizationA",
                "employes": "demo:person3",
            },
        ],
    }
    # graph["@graph"] = sorted(graph["@graph"], key=lambda x: x['@id'])

    context = {
        "schema": "http://schema.org/",
        "demo": "https://oo-ld.github.io/demo/",
        "skos": "http://www.w3.org/2004/02/skos/core#",
        "name": "schema:name",
        "name*": "demo:full_name",
        # "_demo_full_name": "demo:full_name", # generated
        ##"label": {"@id": "skos:prefLabel", "@container": "@set", "@language": "en", "@context": {"text": "@value", "lang": "@language"}},  # noqa
        "text": "@value",
        "lang": "@language",
        "label": {"@id": "skos:prefLabel", "@container": "@set"},
        "label*": {"@id": "demo:label", "@container": "@set", "@language": "en"},
        # "_demo_label": {"@id": "demo:label"},#, "@container": "@set", "@language": "en"}, # generated  # noqa
        "employes": {"@id": "schema:employes", "@type": "@id"},
        "employes*": {"@reverse": "schema:worksFor", "@type": "@id"},
        # "_schema_worksFor": {"@id": "schema:worksFor", "@type": "@id"}, # generated
        "employes**": {"@reverse": "demo:is_employed_by", "@type": "@id"},
        # "_demo_is_employed_by": {"@id": "demo:is_employed_by", "@type": "@id"}, # generated  # noqa
        "type": "@type",
        "id": "@id",
    }

    transformed_graph = jsonld_to_jsonld(graph, context)
    # print("Transformed Graph:", json.dumps(transformed_graph, indent=2))

    expected = {
        "@context": {
            "demo": "https://oo-ld.github.io/demo/",
            "employes": {"@id": "schema:employes", "@type": "@id"},
            "employes*": {"@reverse": "schema:worksFor", "@type": "@id"},
            "employes**": {"@reverse": "demo:is_employed_by", "@type": "@id"},
            "id": "@id",
            "label": {"@container": "@set", "@id": "skos:prefLabel"},
            "label*": {"@container": "@set", "@id": "demo:label", "@language": "en"},
            "lang": "@language",
            "name": "schema:name",
            "name*": "demo:full_name",
            "schema": "http://schema.org/",
            "skos": "http://www.w3.org/2004/02/skos/core#",
            "text": "@value",
            "type": "@type",
        },
        "@graph": [
            {
                "employes": ["demo:person1", "demo:person2", "demo:person3"],
                "id": "demo:organizationA",
                "label": [{"lang": "en", "text": "organizationA"}],
                "type": "schema:Organization",
            },
            {"id": "demo:person1", "name": "Person1", "type": "schema:Person"},
            {"id": "demo:person2", "name": "Person2", "type": "schema:Person"},
            {"id": "demo:person3", "name": "Person3", "type": "schema:Person"},
        ],
    }

    # from jsondiff import diff
    # _diff = json.dumps(diff(transformed_graph, expected), indent=2)
    # assert transformed_graph == expected,
    # f"Expected {expected}, but encountered following deviation: {_diff}"

    assert (
        transformed_graph == expected
    ), f"Expected {expected}, but got {transformed_graph}"


@pytest.mark.skip(reason="This test fails randomly, skip for now")
def test_rocreate():
    rdf = """
        @prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
        @prefix schema: <https://schema.org/> .
        @prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

        <https://example/records/56789/files/139a5c2f-0000-4c04-96a3-04fd4abbf165> a schema:MediaObject ;
            schema:author <https://example/users/1234> ;
            schema:contentSize "177031" ;
            schema:dateCreated "2025-03-20T12:42:16.958207+00:00"^^xsd:dateTime ;
            schema:dateModified "2025-03-20T12:42:17.043920+00:00"^^xsd:dateTime ;
            schema:encodingFormat "image/jpeg" ;
            schema:identifier "139a5c2f-0000-4c04-96a3-04fd4abbf165" ;
            schema:isPartOf <https://example/records/56789> ;
            schema:name "A4.jpg" .

        <https://example/records/56789/files/485d06f8-0000-4804-a615-e29aea4e732b> a schema:MediaObject ;
            schema:author <https://example/users/1234> ;
            schema:contentSize "261300" ;
            schema:dateCreated "2025-03-20T12:42:15.029828+00:00"^^xsd:dateTime ;
            schema:dateModified "2025-03-20T12:42:15.114847+00:00"^^xsd:dateTime ;
            schema:encodingFormat "image/jpeg" ;
            schema:identifier "485d06f8-0000-4804-a615-e29aea4e732b" ;
            schema:isPartOf <https://example/records/56789> ;
            schema:name "C2.jpg" .

        <https://example/records/56789> a schema:Dataset ;
            schema:additionalType "sample" ;
            schema:author <https://example/users/1234> ;
            schema:dateCreated "2025-03-20T12:42:07.127883+00:00"^^xsd:dateTime ;
            schema:dateModified "2025-05-09T08:11:09.848079+00:00"^^xsd:dateTime ;
            schema:identifier "abccell02defects" ;
            schema:keywords "abc02",
                "abccell" ;
            schema:name "ISEcell02defects" .

        <https://example/users/1234> a schema:Person ;
            schema:name "Doe, John" .
    """  # noqa

    g = rdflib.Graph()
    g.parse(data=rdf, format="turtle")
    data = jsonld.expand(
        json.loads(g.serialize(format="json-ld", indent=4, context=None))
    )

    context = {
        "@version": 1.1,  # use version 1.1 for JSON-LD
        "schema": "https://schema.org/",
        "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
        "skos": "http://www.w3.org/2004/02/skos/core#",
        "xsd": "http://www.w3.org/2001/XMLSchema#",
        "type": "@type",  # use @type for type mapping
        "id": "@id",  # use @id for IRI mapping
        # define name as multilingual attribute, using the @language tag (default: en)
        "uuid": {"@id": "schema:identifier"},
        "label": {
            "@id": "skos:prefLabel",
            "@container": "@set",
            "@context": {"text": "@value", "lang": "@language"},
        },
        "label*": {"@id": "schema:name", "@language": "en"},
        "creation_date": {"@id": "schema:dateCreated", "@type": "xsd:dateTime"},
        "edition_date": {"@id": "schema:dateModified", "@type": "xsd:dateTime"},
        "encoding_format": {"@id": "schema:encodingFormat"},
        "content_size": {"@id": "schema:contentSize"},
        "WikiFile": {"@id": "Category:WikiFile", "@type": "@id"},
        "WikiFile*": {"@id": "schema:MediaObject", "@type": "@id"},
        "ELNEntry": {"@id": "Category:ELNEntry", "@type": "@id"},
        "ELNEntry*": {"@id": "schema:Dataset", "@type": "@id"},
        "creator": {"@id": "schema:author", "@type": "@id"},
        "attachments": {"@id": "Property:HasFileAttachment", "@type": "@id"},
        "attachments*": {"@reverse": "schema:isPartOf", "@type": "@id"},
        "keywords": {"@id": "schema:keywords", "@container": "@set"},
    }

    transformed_graph = jsonld_to_jsonld(data, context)

    expected = [
        {
            "id": "https://example/records/56789",
            "type": "ELNEntry",
            "attachments": [
                "https://example/records/56789/files/485d06f8-0000-4804-a615-e29aea4e732b",  # noqa
                "https://example/records/56789/files/139a5c2f-0000-4c04-96a3-04fd4abbf165",  # noqa
            ],
            "label": [{"lang": "en", "text": "ISEcell02defects"}],
            "schema:additionalType": "sample",
            "creator": "https://example/users/1234",
            "creation_date": "2025-03-20T12:42:07.127883+00:00",
            "edition_date": "2025-05-09T08:11:09.848079+00:00",
            "uuid": "abccell02defects",
            "keywords": ["abc02", "abccell"],
        },
        {
            "id": "https://example/records/56789/files/139a5c2f-0000-4c04-96a3-04fd4abbf165",  # noqa
            "type": "WikiFile",
            "label": [{"lang": "en", "text": "A4.jpg"}],
            "creator": "https://example/users/1234",
            "content_size": "177031",
            "creation_date": "2025-03-20T12:42:16.958207+00:00",
            "edition_date": "2025-03-20T12:42:17.043920+00:00",
            "encoding_format": "image/jpeg",
            "uuid": "139a5c2f-0000-4c04-96a3-04fd4abbf165",
        },
        {
            "id": "https://example/records/56789/files/485d06f8-0000-4804-a615-e29aea4e732b",  # noqa
            "type": "WikiFile",
            "label": [{"lang": "en", "text": "C2.jpg"}],
            "creator": "https://example/users/1234",
            "content_size": "261300",
            "creation_date": "2025-03-20T12:42:15.029828+00:00",
            "edition_date": "2025-03-20T12:42:15.114847+00:00",
            "encoding_format": "image/jpeg",
            "uuid": "485d06f8-0000-4804-a615-e29aea4e732b",
        },
        {
            "id": "https://example/users/1234",
            "type": "schema:Person",
            "label": [{"lang": "en", "text": "Doe, John"}],
        },
    ]

    _diff = diff(transformed_graph["@graph"], expected)
    assert (
        transformed_graph["@graph"] == expected
    ), f"Expected {expected}, but encountered following deviation: {_diff}"


@pytest.mark.skip(
    reason="This test fails due to the literal type annotation issue in JSON-LD"
)
def test_literal_types():
    data = {
        "https://schema.org/contentSize": 177031,
    }
    context = {
        "@version": 1.1,
        "schema": "https://schema.org/",
        "xsd": "http://www.w3.org/2001/XMLSchema#",
        "content_size": {"@id": "schema:contentSize", "@type": "xsd:integer"},
    }

    expected = {
        "@context": {
            "@version": 1.1,
            "schema": "https://schema.org/",
            "xsd": "http://www.w3.org/2001/XMLSchema#",
            "content_size": {"@id": "schema:contentSize", "@type": "xsd:integer"},
        },
        "content_size": 177031,
    }
    data = jsonld.compact(data, context)
    print(json.dumps(data, indent=4))
    # this fails due to the literal type annotation
    # instead we get "schema:contentSize": 177031
    # see also https://github.com/json-ld/json-ld.org/issues/795
    assert data == expected, f"Expected {expected}, but got {data}"


if __name__ == "__main__":
    test_simple_json()
    test_complex_graph()
    test_rocreate()
    test_literal_types()



import json

from pyld import jsonld


def jsonld_to_jsonld(graph: dict, transformation_context: dict) -> dict:
    """Applies OO-LD alias notation to transform JSON(-LD) documents

    Parameters
    ----------
    graph
        input JSON(-LD) document to transform
    context
        transformation context, which contains the mapping of aliases to URIs

    Returns
    -------
        transformed JSON(-LD) document
    """
    temp1 = {}
    temp2 = {}
    temp3 = {}

    # # not expanded => at least one key is not an IRI
    # expanded = True

    # if "@context" in graph: expanded = False
    # elif "@graph" in graph:
    #     for item in graph["@graph"]:
    #         if not is_iri(item):
    #             expanded = False
    #             break
    # else:
    #     for key in graph.keys():
    #         if not (":" in value):
    #             expanded = False
    #             break

    # if not context is given, we assume the default context
    # if '@context' not in graph:
    #    graph["@context"] = context

    # in case graph is not expanded we expand first
    # graph = jsonld.expand(graph)

    for key, value in transformation_context.items():
        if key.endswith("*"):
            temp1_value = {}
            temp2_value = {}
            if type(value) is dict:
                if "@id" in value:
                    temp1_value["@id"] = value["@id"]
                if "@reverse" in value:
                    temp1_value["@id"] = value["@reverse"]
                if "@type" in value:
                    temp1_value["@type"] = value["@type"]
                temp2_value = {**value}
                # if "@id" in value: del temp2_value["@id"]
                # if "@reverse" in value: del temp2_value["@reverse"]
            else:
                temp1_value["@id"] = value
                temp2_value["@id"] = value

            org_key = key.replace("*", "")
            org_value = transformation_context[org_key]
            if type(org_value) is dict:
                if "@id" in org_value:
                    # temp2_value["@id"] = org_value["@id"]
                    if "@id" in temp2_value:
                        temp2_value["@id"] = org_value["@id"]
                    if "@reverse" in temp2_value:
                        temp2_value["@reverse"] = org_value["@id"]
                # if "@reverse" in org_value: temp2_value["@id"] = org_value["@reverse"]
                else:
                    print("Error")
            else:
                if "@id" in temp2_value:
                    temp2_value["@id"] = org_value
                if "@reverse" in temp2_value:
                    temp2_value["@reverse"] = org_value

            temp1["_" + temp1_value["@id"].replace(":", "_")] = temp1_value
            temp2["_" + temp1_value["@id"].replace(":", "_")] = temp2_value

            # asume type mapping if key starts with capital letter
            if key[0].isupper():
                temp1[key] = None
                # temp3[org_key] = None

    print("temp1", temp1)
    print("temp2", temp2)
    print("temp3", temp3)
    graph = jsonld.compact(graph, {**transformation_context, **temp1})

    graph["@context"] = {**transformation_context, **temp2}
    graph = jsonld.flatten(graph)  # may introduce blank node @ids
    # graph = jsonld.expand(graph) # does not resolve @reverse relations

    graph = jsonld.compact(graph, {**transformation_context, **temp3})

    return graph


def json_to_json(
    document: dict, transformation_context: dict, document_context: dict = None
) -> dict:
    """Applies OO-LD alias notation to transform JSON documents

    Parameters
    ----------
    document
        input JSON document to transform
    transformation_context
        transformation context, which contains the mapping of aliases to URIs
    document_context
        context of the input document. Defaults to transformation_context

    Returns
    -------
        transformed JSON document
    """
    if document_context is None:
        document_context = transformation_context
    if type(document_context) is list:
        document = {"@graph": document}
    document["@context"] = document_context
    document = jsonld.expand(document)
    # anonymous_document = "@id" not in document
    # if "@graph" in document:
    #    # check none of the items in the graph has an @id
    #    anonymous_document = all("@id" not in item for item in document["@graph"])
    # check if json-string contains an @id
    anonymous_document = "@id" not in json.dumps(document)
    document = jsonld_to_jsonld(document, transformation_context)

    if anonymous_document:
        context = document["@context"]
        document = jsonld.expand(document)

        # if "@id" in document:
        #     del document["@id"]
        # if "@graph" in document:
        #     for item in document["@graph"]:
        #         if "@id" in item:
        #             del item["@id"]
        # recursively delete all @id keys in the document
        def remove_ids(d):
            if isinstance(d, dict):
                d.pop("@id", None)
                for key in list(d.keys()):
                    remove_ids(d[key])
            elif isinstance(d, list):
                for item in d:
                    remove_ids(item)

        remove_ids(document)
        document = jsonld.compact(document, context)
    del document["@context"]
    return document



from enum import Enum
from typing import List, Optional

import panel as pn
from pydantic import ConfigDict, Field

from oold.model import LinkedBaseModel
from oold.ui.panel.demo import OoldDemoEditor


class Entity(LinkedBaseModel):
    model_config = ConfigDict(
        json_schema_extra={
            "@context": {
                # aliases
                "id": "@id",
                "type": "@type",
                # prefixes
                "schema": "https://schema.org/",
                "ex": "https://example.com/",
                # literal property
                "name": "schema:name",
            },
            "iri": "Entity.json",  # the IRI of the schema
        }
    )
    type: Optional[str] = Field(
        "ex:Entity.json",
        json_schema_extra={"options": {"hidden": "true"}},
    )
    name: str

    def get_iri(self):
        return "ex:" + self.name


class Hobby(str, Enum):
    """Various hobbies as an enum."""

    SPORTS = "ex:sports"
    """Sports hobby, e.g. football, basketball, etc."""
    MUSIC = "ex:music"
    """Music hobby, e.g. playing instruments, singing, etc."""
    ART = "ex:art"
    """Art hobby, e.g. painting, drawing, etc."""


class Person(Entity):
    """A simple Person schema"""

    model_config = ConfigDict(
        json_schema_extra={
            "@context": [
                "Entity.json",  # import the context of the parent class
                {
                    # object property definition
                    "hobbies": {
                        "@id": "ex:hobbies",
                        "@type": "@id",
                    },
                    "knows": {
                        "@id": "schema:knows",
                        "@type": "@id",
                        "@container": "@set",
                    },
                },
            ],
            "iri": "Person.json",
            "defaultProperties": ["type", "name", "hobbies"],
        }
    )
    type: Optional[str] = "ex:Person.json"
    hobbies: Optional[List[Hobby]] = None
    """interests of the person, e.g. sports, music, art"""
    knows: Optional[List["Person"]] = Field(
        None,
        # object property pointing to another Person
        json_schema_extra={"range": "Person.json"},
    )


editor = OoldDemoEditor(Person)
pn.serve(editor.servable())



import json
from pprint import pprint
from typing import Dict, Optional, Union

from pydantic import BaseModel, Field


class LinkedBaseModel(BaseModel):
    type: str
    id: str
    __iris__: Optional[Dict[str, str]] = {}

    def __init__(self, *a, **kw):
        # pprint(a)
        # pprint(kw)
        for name in list(kw):  # force copy of keys for inline-delete
            # rewrite <attr> to <attr>_iri
            # pprint(self.__fields__)
            # if hasattr(self.__fields__[name], "extra") and "range" in self.__fields__[name].extra: # pydantic v1
            if not "__iris__" in kw:
                kw["__iris__"] = {}
            if (
                self.model_fields[name].json_schema_extra
                and "range" in self.model_fields[name].json_schema_extra
            ):  # pydantic v2: model_fields
                if isinstance(kw[name], BaseModel):  # contructed with object ref
                    # print(kw[name].id)
                    kw["__iris__"][name] = kw[name].id
                elif isinstance(kw[name], str):  # constructed from json
                    kw["__iris__"][name] = kw[name]
                    del kw[name]
        pprint(kw)
        super().__init__(*a, **kw)
        self.__iris__ = kw["__iris__"]

    def __getattribute__(self, name):
        # print("__getattribute__ ", name)
        # async? https://stackoverflow.com/questions/33128325/how-to-set-class-attribute-with-await-in-init
        if name in ["__dict__", "__pydantic_private__", "__iris__"]:
            return BaseModel.__getattribute__(self, name)  # prevent loop
        # if name in ["__pydantic_extra__"]
        if "__iris__" in self.__dict__:
            if name in self.__dict__["__iris__"]:
                # if self.__iris__:
                #    if name in self.__iris__:
                # print("in dict")
                iri = self.__iris__[name]
                # we will need an osw instance here
                # if iri in graph:
                # print("in graph")
                node = get(iri)
                if node:
                    self.__setattr__(name, node)
        return BaseModel.__getattribute__(self, name)

    def _object_to_iri(self, d):
        for name in list(d):  # force copy of keys for inline-delete
            if name in self.__iris__:
                d[name] = self.__iris__[name]
                # del d[name + "_iri"]
        return d

    def dict(self, **kwargs):  # extent BaseClass export function
        print("dict")
        d = super().dict(**kwargs)
        # pprint(d)
        self._object_to_iri(d)
        # pprint(d)
        return d

    def json(self, **kwargs):
        # print("json")
        d = json.loads(BaseModel.json(self, **kwargs))  # ToDo directly use dict?
        self._object_to_iri(d)
        return json.dumps(d, **kwargs)

    def model_dump_json(self, **kwargs):
        # print("json")
        d = json.loads(
            BaseModel.model_dump_json(self, **kwargs)
        )  # ToDo directly use dict?
        self._object_to_iri(d)
        return json.dumps(d, **kwargs)


class Bar(LinkedBaseModel):
    type: Optional[str] = "Bar"
    prop1: str


class Foo(LinkedBaseModel):
    type: Optional[str] = "Foo"
    literal: str
    b: Optional[Bar] = Field(None, range="ex:test")


graph = [
    {
        "id": "ex:LinkedBaseModel",
        "type": "class",
        "properties": {"type": {"type": "string"}},
    },
    {
        "id": "ex:Bar",
        "type": "class",
        "properties": {"type": {"default": "Bar"}, "prop1": {"type": "string"}},
    },
    {
        "id": "ex:Foo",
        "type": "class",
        "properties": {
            "type": {"default": "Foo"},
            "literal": {"type": "string"},
            "b": {"type": "string", "format": "iri", "range": "ex:Bar"},
        },
    },
    {"id": "ex:a", "type": "Foo", "literal": "test1", "b": "ex:b"},
    {"id": "ex:b", "type": "Bar", "prop1": "test2"},
]


def get(iri) -> Union[None, LinkedBaseModel]:
    for node in graph:
        if node["id"] == iri:
            cls = node["type"]
            entity = eval(f"{cls}(**node)")
            return entity
    return None


b = Bar(id="ex:b1", prop1="test")
f2 = Foo(id="ex:f2", literal="test1", b=b)
f = get("ex:a")
pprint(f)
# print(f.b_iri)
pprint(f.b)
# pydantic 1
# pprint(f.json())
# pprint(f2.json())
# pydantic 2
pprint(f.model_dump_json())
pprint(f2.model_dump_json())



# generated by datamodel-codegen:
#   filename:  Foo.json
#   timestamp: 2025-06-16T05:00:41+00:00

from __future__ import annotations

from typing import List, Optional

from pydantic.v1 import Field

from oold.model.v1 import LinkedBaseModel


class Bar2(LinkedBaseModel):
    class Config:
        schema_extra = {"title": "Bar2"}

    id: Optional[str] = None
    type: Optional[List[str]] = ["Bar2"]
    prop1: Optional[str] = None


class Bar(Bar2):
    class Config:
        schema_extra = {"title": "Bar"}

    type: Optional[List[str]] = ["Bar"]
    prop2: Optional[str] = None


class Foo(LinkedBaseModel):
    class Config:
        schema_extra = {"title": "Foo"}

    id: str
    type: Optional[List[str]] = ["Foo"]
    literal: Optional[str] = None
    b: Optional[Bar] = Field(None, range="Bar.json", x_oold_required_iri=True)
    b_default: Optional[Bar] = Field(
        default_factory=lambda: Bar.parse_obj("ex:b"), range="Bar.json"
    )
    b_set_later: Optional[Bar] = Field(None, range="Bar.json")
    b2: Optional[List[Bar]] = Field(None, range="Bar.json")
